#!usr/bin/env python
# -*- coding: utf-8 -*-

'''
Authors: 
  Sebastian Burgstaller (sebastian.burgstaller' at 'gmail.com
  Andra Waagmeester (andra' at ' micelio.be)

This file is part of ProteinBoxBot.

ProteinBoxBot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ProteinBoxBot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProteinBoxBot.  If not, see <http://www.gnu.org/licenses/>.
'''


# from raven import Client

def getSentryKey():
    return '<Sentry key to be obtained from https://getsentry.com/welcome/>'


def getWikiDataUser():
    return 'ProteinBoxBot'


def getWikiDataPassword():
    return '<WikiData Password>'


def getMySQLHost():
    return '127.0.0.1'


def getMySQLUser():
    return 'PBB'


def getMySQLPW():
    return ''
